@@ -70,7 +70,7 @@ class ArgType:
 	parameter."""
 	raise RuntimeError, "write_param not implemented for %s" % \
               self.__class__.__name__
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
 	"""Adds a variable named ret of the return type to
 	info.varlist, and add any required code to info.codeafter to
 	convert the return value to a python object."""
@@ -78,7 +78,7 @@ class ArgType:
               self.__class__.__name__
 
 class NoneArg(ArgType):
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.codeafter.append('    Py_INCREF(Py_None);\n' +
                               '    return Py_None;')
 
@@ -94,14 +94,8 @@ class StringArg(ArgType):
             info.add_parselist('z', ['&' + pname], [pname])
 	else:
             info.add_parselist('s', ['&' + pname], [pname])
-    def write_return(self, ptype, info):
-        if ptype[:6] == 'const-' or ptype in ('static_string',):
-	    info.varlist.add('const gchar', '*ret')
-            info.codeafter.append('    if (ret)\n' +
-                                  '        return PyString_FromString(ret);\n'+
-                                  '    Py_INCREF(Py_None);\n' +
-                                  '    return Py_None;')
-	else:
+    def write_return(self, ptype, ownsreturn, info):
+        if ownsreturn:
 	    # have to free result ...
 	    info.varlist.add('gchar', '*ret')
             info.codeafter.append('    if (ret) {\n' +
@@ -111,6 +105,12 @@ class StringArg(ArgType):
                                   '    }\n' +
                                   '    Py_INCREF(Py_None);\n' +
                                   '    return Py_None;')
+	else:
+	    info.varlist.add('const gchar', '*ret')
+            info.codeafter.append('    if (ret)\n' +
+                                  '        return PyString_FromString(ret);\n'+
+                                  '    Py_INCREF(Py_None);\n' +
+                                  '    return Py_None;')
 
 class UCharArg(ArgType):
     # allows strings with embedded NULLs.
@@ -128,24 +128,6 @@ class UCharArg(ArgType):
             info.add_parselist('s#', ['&' + pname, '&' + pname + '_len'],
                                [pname])
 
-    def write_return(self, ptype, info):
-        if ptype.startswith('const-'):
-	    info.varlist.add('const guchar', '*ret')
-            info.codeafter.append('    if (ret)\n' +
-                                  '        return PyUnicode_DeviceUTF8(ret, strlen(ret), "strict");\n'+
-                                  '    Py_INCREF(Py_None);\n' +
-                                  '    return Py_None;')
-	else:
-	    # have to free result ...
-	    info.varlist.add('guchar', '*ret')
-            info.codeafter.append('    if (ret) {\n' +
-                                  '        PyObject *py_ret = PyUnicode_DecodeUTF8(ret, strlen(ret), "strict");\n' +
-                                  '        g_free(ret);\n' +
-                                  '        return py_ret;\n' +
-                                  '    }\n' +
-                                  '    Py_INCREF(Py_None);\n' +
-                                  '    return Py_None;')
-
 class CharArg(ArgType):
     def write_param(self, ptype, pname, pdflt, pnull, info):
 	if pdflt:
@@ -154,9 +136,45 @@ class CharArg(ArgType):
 	    info.varlist.add('char', pname)
 	info.arglist.append(pname)
         info.add_parselist('c', ['&' + pname], [pname])
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
 	info.varlist.add('gchar', 'ret')
         info.codeafter.append('    return PyString_FromStringAndSize(&ret, 1);')
+class GUniCharArg(ArgType):
+    param_tmpl = ('    if (py_%(name)s[1] != 0) {\n'
+                  '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a 1 character unicode string");\n'
+                  '        return NULL;\n'
+                  '    }\n'
+                  '    %(name)s = (gunichar)py_%(name)s[0];\n')
+    dflt_tmpl = ('    if (py_%(name)s != NULL) {\n'
+                 '        if (py_%(name)s[1] != 0) {\n'
+                 '            PyErr_SetString(PyExc_TypeError, "%(name)s should be a 1 character unicode string");\n'
+                 '            return NULL;\n'
+                 '        }\n'
+                 '        %(name)s = (gunichar)py_%(name)s[0];\n'
+                 '     }\n')
+    ret_tmpl = ('#if !defined(Py_UNICODE_SIZE) || Py_UNICODE_SIZE == 2\n'
+                '    if (ret > 0xffff) {\n'
+                '        PyErr_SetString(PyExc_RuntimeError, "returned character can not be represented in 16-bit unicode");\n'
+                '        return NULL;\n'
+                '    }\n'
+                '#endif\n'
+                '    py_ret = (Py_UNICODE)ret;\n'
+                '    return PyUnicode_FromUnicode(&py_ret, 1);\n')
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+	if pdflt:
+	    info.varlist.add('gunichar', pname + " = '" + pdflt + "'")
+            info.codebefore.append(self.dflt_tmpl % {'name':pname})
+	else:
+	    info.varlist.add('gunichar', pname)
+            info.codebefore.append(self.param_tmpl % {'name':pname})
+        info.varlist.add('Py_UNICODE', '*py_' + pname + ' = NULL')
+	info.arglist.append(pname)
+        info.add_parselist('u', ['&py_' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add('gunichar', 'ret')
+        info.varlist.add('Py_UNICODE', 'py_ret')
+        info.codeafter.append(self.ret_tmpl)
+        
 
 class IntArg(ArgType):
     def write_param(self, ptype, pname, pdflt, pnull, info):
@@ -166,10 +184,59 @@ class IntArg(ArgType):
 	    info.varlist.add('int', pname)
 	info.arglist.append(pname)
         info.add_parselist('i', ['&' + pname], [pname])
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('int', 'ret')
         info.codeafter.append('    return PyInt_FromLong(ret);')
 
+class ULongArg(ArgType):
+    dflt = '    if (py_%(name)s)\n' \
+           '        %(name)s = PyLong_AsUnsignedLong(py_%(name)s);\n'
+    before = '    %(name)s = PyLong_AsUnsignedLong(py_%(name)s);\n'
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+        if pdflt:
+            info.varlist.add('gulong', pname + ' = ' + pdflt)
+            info.codebefore.append(self.dflt % {'name':pname})            
+        else:
+            info.varlist.add('gulong', pname)
+            info.codebefore.append(self.before % {'name':pname})            
+        info.varlist.add('PyObject', "*py_" + pname + ' = NULL')
+        info.arglist.append(pname)
+        info.add_parselist('O!', ['&PyLong_Type', '&py_' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add('gulong', 'ret')
+        info.codeafter.append('    return PyLong_FromUnsignedLong(ret);')
+
+class Int64Arg(ArgType):
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+	if pdflt:
+	    info.varlist.add('gint64', pname + ' = ' + pdflt)
+	else:
+	    info.varlist.add('gint64', pname)
+	info.arglist.append(pname)
+        info.add_parselist('L', ['&' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add('gint64', 'ret')
+        info.codeafter.append('    return PyLong_FromLongLong(ret);')
+
+class UInt64Arg(ArgType):
+    dflt = '    if (py_%(name)s)\n' \
+           '        %(name)s = PyLong_AsUnsignedLongLong(py_%(name)s);\n'
+    before = '    %(name)s = PyLong_AsUnsignedLongLong(py_%(name)s);\n'
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+        if pdflt:
+            info.varlist.add('guint64', pname + ' = ' + pdflt)
+            info.codebefore.append(self.dflt % {'name':pname})            
+        else:
+            info.varlist.add('guint64', pname)
+            info.codebefore.append(self.before % {'name':pname})            
+        info.varlist.add('PyObject', "*py_" + pname + ' = NULL')
+        info.arglist.append(pname)
+        info.add_parselist('O!', ['&PyLong_Type', '&py_' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        info.varlist.add('guint64', 'ret')
+        info.codeafter.append('    return PyLong_FromUnsignedLongLong(ret);')
+        
+
 class DoubleArg(ArgType):
     def write_param(self, ptype, pname, pdflt, pnull, info):
 	if pdflt:
@@ -178,7 +245,7 @@ class DoubleArg(ArgType):
 	    info.varlist.add('double', pname)
 	info.arglist.append(pname)
         info.add_parselist('d', ['&' + pname], [pname])
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('double', 'ret')
         info.codeafter.append('    return PyFloat_FromDouble(ret);')
 
@@ -221,7 +288,7 @@ class FileArg(ArgType):
 		info.varlist.add('PyObject', '*' + pname)
 		info.arglist.append('PyFile_AsFile(' + pname + ')')
             info.add_parselist('O!', ['&PyFile_Type', '&' + pname], [pname])
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
 	info.varlist.add('FILE', '*ret')
         info.codeafter.append('    if (ret)\n' +
                               '        return PyFile_FromFile(ret, "", "", fclose);\n' +
@@ -244,7 +311,7 @@ class EnumArg(ArgType):
                                              'name': pname})
 	info.arglist.append(pname)
         info.add_parselist('O', ['&py_' + pname], [pname]);
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('gint', 'ret')
         info.codeafter.append('    return PyInt_FromLong(ret);')
 
@@ -267,7 +334,7 @@ class FlagsArg(ArgType):
                                             'name':pname})
 	info.arglist.append(pname)
         info.add_parselist('O', ['&py_' + pname], [pname])
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('guint', 'ret')
         info.codeafter.append('    return PyInt_FromLong(ret);')
 
@@ -288,16 +355,8 @@ class ObjectArg(ArgType):
             '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a %(type)s or None");\n'
             '        return NULL;\n'
             '    }\n')
-    dflt = ('    if (py_%(name)s && pygobject_check(py_%(name)s, &Py%(type)s_Type))\n'
-            '        %(name)s = %(cast)s(py_%(name)s->obj);\n'
-            '    else if (py_%(name)s) {\n'
-            '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a %(type)s");\n'
-            '        return NULL;\n'
-            '    }\n')
-    check = ('    if (!pygobject_check(%(name)s, &Py%(type)s_Type)) {\n'
-             '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a %(type)s");\n'
-             '        return NULL;\n'
-             '    }\n')
+    dflt = '    if (py_%(name)s)\n' \
+           '        %(name)s = %(cast)s(py_%(name)s->obj);\n'
     def __init__(self, objname, parent, typecode):
 	self.objname = objname
 	self.cast = string.replace(typecode, '_TYPE_', '_', 1)
@@ -323,22 +382,26 @@ class ObjectArg(ArgType):
 		info.varlist.add(self.objname, '*' + pname + ' = ' + pdflt)
 		info.varlist.add('PyGObject', '*py_' + pname + ' = NULL')
 		info.codebefore.append(self.dflt % {'name':pname,
-                                                    'cast':self.cast,
-                                                    'type':self.objname}) 
+                                                    'cast':self.cast}) 
 		info.arglist.append(pname)
-                info.add_parselist('O', ['&py_' + pname], [pname])
+                info.add_parselist('O', ['&Py%s_Type' % self.objname,
+                                         '&py_' + pname], [pname])
 	    else:
 		info.varlist.add('PyGObject', '*' + pname)
-		info.codebefore.append(self.check % {'name':pname,
-                                                     'cast':self.cast,
-                                                     'type':self.objname}) 
 		info.arglist.append('%s(%s->obj)' % (self.cast, pname))
-                info.add_parselist('O', ['&' + pname], [pname])
-    def write_return(self, ptype, info):
+                info.add_parselist('O!', ['&Py%s_Type' % self.objname,
+                                          '&' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
         if ptype[-1] == '*': ptype = ptype[:-1]
         info.varlist.add(ptype, '*ret')
-        info.codeafter.append('    /* pygobject_new handles NULL checking */\n' +
-                              '    return pygobject_new((GObject *)ret);')
+        if ownsreturn:
+            info.varlist.add('PyObject', '*py_ret')
+            info.codeafter.append('    py_ret = pygobject_new((GObject *)ret;\n'
+                                  '    g_object_unref(ret);\n'
+                                  '    return py_ret;')
+        else:
+            info.codeafter.append('    /* pygobject_new handles NULL checking */\n' +
+                                  '    return pygobject_new((GObject *)ret);')
 
 class BoxedArg(ArgType):
     # haven't done support for default args.  Is it needed?
@@ -370,17 +433,30 @@ class BoxedArg(ArgType):
 	    info.codebefore.append(self.check % {'name':  pname,
                                                  'typename': self.typename,
                                                  'typecode': self.typecode})
-        info.arglist.append(pname)
+        if ptype[-1] == '*':
+            typename = ptype[:-1]
+            if typename[:6] == 'const-': typename = typename[6:]
+            if typename != self.typename:
+                info.arglist.append('(%s *)%s' % (ptype[:-1], pname))
+            else:
+                info.arglist.append(pname)
+        else:
+            info.arglist.append(pname)
         info.add_parselist('O', ['&py_' + pname], [pname])
-    def write_return(self, ptype, info):
+    ret_tmpl = '    /* pyg_boxed_new handles NULL checking */\n' \
+               '    return pyg_boxed_new(%(typecode)s, %(ret)s, %(copy)s, TRUE);'
+    def write_return(self, ptype, ownsreturn, info):
         if ptype[-1] == '*':
             info.varlist.add(self.typename, '*ret')
-            info.codeafter.append('    /* pyg_boxed_new handles NULL checking */\n' +
-                                  '    return pyg_boxed_new(' + self.typecode + ', ret, TRUE, TRUE);')
+            ret = 'ret'
         else:
             info.varlist.add(self.typename, 'ret')
-            info.codeafter.append('    /* pyg_boxed_new handles NULL checking */\n' +
-                                  '    return pyg_boxed_new(' + self.typecode + ', &ret, TRUE, TRUE);')
+            ret = '&ret'
+            ownsreturn = 0 # of course it can't own a ref to a local var ...
+        info.codeafter.append(self.ret_tmpl %
+                              { 'typecode': self.typecode,
+                                'ret': ret,
+                                'copy': ownsreturn and 'FALSE' or 'TRUE'})
 
 class CustomBoxedArg(ArgType):
     # haven't done support for default args.  Is it needed?
@@ -409,13 +485,55 @@ class CustomBoxedArg(ArgType):
 	    info.varlist.add('PyObject', '*' + pname)
 	    info.arglist.append(self.getter + '(' + pname + ')')
             info.add_parselist('O!', ['&' + self.pytype, '&' + pname], [pname])
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.varlist.add(ptype[:-1], '*ret')
         info.codeafter.append('    if (ret)\n' +
                               '        return ' + self.new + '(ret);\n' +
                               '    Py_INCREF(Py_None);\n' +
                               '    return Py_None;')
 
+class PointerArg(ArgType):
+    # haven't done support for default args.  Is it needed?
+    check = ('    if (pyg_pointer_check(py_%(name)s, %(typecode)s))\n'
+             '        %(name)s = pyg_pointer_get(py_%(name)s, %(typename)s);\n'
+             '    else {\n'
+             '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a %(typename)s");\n'
+             '        return NULL;\n'
+             '    }\n')
+    null = ('    if (pyg_pointer_check(py_%(name)s, %(typecode)s))\n'
+            '        %(name)s = pyg_pointer_get(py_%(name)s, %(typename)s);\n'
+            '    else if (py_%(name)s != Py_None) {\n'
+            '        PyErr_SetString(PyExc_TypeError, "%(name)s should be a %(typename)s or None");\n'
+            '        return NULL;\n'
+            '    }\n')
+    def __init__(self, ptype, typecode):
+	self.typename = ptype
+	self.typecode = typecode
+    def write_param(self, ptype, pname, pdflt, pnull, info):
+	if pnull:
+            info.varlist.add(self.typename, '*' + pname + ' = NULL')
+	    info.varlist.add('PyObject', '*py_' + pname + ' = Py_None')
+	    info.codebefore.append(self.null % {'name':  pname,
+                                                'typename': self.typename,
+                                                'typecode': self.typecode})
+	else:
+            info.varlist.add(self.typename, '*' + pname + ' = NULL')
+	    info.varlist.add('PyObject', '*py_' + pname)
+	    info.codebefore.append(self.check % {'name':  pname,
+                                                 'typename': self.typename,
+                                                 'typecode': self.typecode})
+        info.arglist.append(pname)
+        info.add_parselist('O', ['&py_' + pname], [pname])
+    def write_return(self, ptype, ownsreturn, info):
+        if ptype[-1] == '*':
+            info.varlist.add(self.typename, '*ret')
+            info.codeafter.append('    /* pyg_pointer_new handles NULL checking */\n' +
+                                  '    return pyg_pointer_new(' + self.typecode + ', ret);')
+        else:
+            info.varlist.add(self.typename, 'ret')
+            info.codeafter.append('    /* pyg_pointer_new handles NULL checking */\n' +
+                                  '    return pyg_pointer_new(' + self.typecode + ', &ret);')
+
 class AtomArg(IntArg):
     atom = ('    %(name)s = pygdk_atom_from_pyobject(py_%(name)s);\n'
             '    if (PyErr_Occurred())\n'
@@ -426,7 +544,7 @@ class AtomArg(IntArg):
 	info.codebefore.append(self.atom % {'name': pname})
 	info.arglist.append(pname)
         info.add_parselist('O', ['&py_' + pname], [pname])
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('GdkAtom', 'ret')
         info.codeafter.append('    return PyGdkAtom_New(ret);')
 
@@ -439,7 +557,7 @@ class GTypeArg(ArgType):
 	info.codebefore.append(self.gtype % {'name': pname})
 	info.arglist.append(pname)
         info.add_parselist('O', ['&py_' + pname], [pname])
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('GType', 'ret')
         info.codeafter.append('    return pyg_type_wrapper_new(ret);')
 
@@ -454,7 +572,18 @@ class GErrorArg(ArgType):
 
 class GtkTreePathArg(ArgType):
     # haven't done support for default args.  Is it needed?
-    normal = '    %(name)s = pygtk_tree_path_from_pyobject(py_%(name)s);\n'
+    normal = ('    %(name)s = pygtk_tree_path_from_pyobject(py_%(name)s);\n'
+              '    if (!%(name)s) {\n'
+              '        PyErr_SetString(PyExc_TypeError, "could not convert %(name)s to a GtkTreePath");\n'
+              '        return NULL;\n'
+              '    }\n')
+    null = ('    if (py_%(name)s != Py_None) {\n'
+            '        %(name)s = pygtk_tree_path_from_pyobject(py_%(name)s);\n'
+            '        if (!%(name)s) {\n'
+            '            PyErr_SetString(PyExc_TypeError, "could not convert %(name)s to a GtkTreePath");\n'
+            '            return NULL;\n'
+            '        }\n'
+            '    }\n')
     null = ('    if (PyTuple_Check(py_%(name)s))\n'
             '        %(name)s = pygtk_tree_path_from_pyobject(py_%(name)s);\n'
             '    else if (py_%(name)s != Py_None) {\n'
@@ -473,50 +602,45 @@ class GtkTreePathArg(ArgType):
 	    info.arglist.append(pname)
             info.add_parselist('O', ['&py_' + pname], [pname])
 	else:
-            info.varlist.add('GtkTreePath', '*' + pname + ' = NULL')
+            info.varlist.add('GtkTreePath', '*' + pname)
 	    info.varlist.add('PyObject', '*py_' + pname)
             info.codebefore.append(self.normal % {'name': pname})
 	    info.arglist.append(pname)
-            info.add_parselist('O!', ['&PyTuple_Type', '&py_' + pname],[pname])
+            info.add_parselist('O', ['&py_' + pname], [pname])
         info.codeafter.append(self.freepath % {'name': pname})
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
         info.varlist.add('GtkTreePath', '*ret')
-        info.codeafter.append('    if (ret) {\n' +
-                              '        PyObject *py_ret = pygtk_tree_path_to_pyobject(ret);\n' +
-                              '        gtk_tree_path_free(ret);\n' +
-                              '        return py_ret;\n' +
-                              '    }\n' +
-                              '    Py_INCREF(Py_None);\n' +
-                              '    return Py_None;')
-													       
+        if ownsreturn:
+            info.codeafter.append('    if (ret) {\n'
+                                  '        PyObject *py_ret = pygtk_tree_path_to_pyobject(ret);\n'
+                                  '        gtk_tree_path_free(ret);\n'
+                                  '        return py_ret;\n'
+                                  '    }\n'
+                                  '    Py_INCREF(Py_None);\n'
+                                  '    return Py_None;')
+        else:
+            info.codeafter.append('    if (ret) {\n'
+                                  '        PyObject *py_ret = pygtk_tree_path_to_pyobject(ret);\n'
+                                  '        return py_ret;\n'
+                                  '    }\n'
+                                  '    Py_INCREF(Py_None);\n'
+                                  '    return Py_None;')
 						       
 class GdkRectanglePtrArg(ArgType):
-    null = ('    if (PyTuple_Check(py_%(name)s)) {\n'
-            '        %(name)s = g_new(GdkRectangle, 1);\n'
-	    '        if (!PyArg_ParseTuple(py_%(name)s, "iiii", &%(name)s->x, &%(name)s->y, &%(name)s->width, &%(name)s->height)) {\n'
-	    '            return NULL;\n'
-	    '        }\n'
-	    '    } else {\n'
-	    '        PyErr_SetString(PyExc_TypeError, "%(name)s must be a tuple");\n'
-	    '    }\n')
+    null = ('    if (!pygdk_rectangle_from_pyobject(py_%(name)s, &%(name)s))\n'
+            '        return NULL;\n')
     def write_param(self, ptype, pname, pdflt, pnull, info):
-	if pnull:
-	    info.varlist.add('GdkRectangle', '*' + pname + ' = NULL')
-	    info.varlist.add('PyObject', '*py_' + pname + ' = Py_None')
-            info.add_parselist('O', ['&py_' + pname], [pname])
-	else:
-	    info.varlist.add('GdkRectangle', '*' + pname + ' = NULL')
-	    info.varlist.add('PyObject', '*py_' + pname)
-            info.add_parselist('O!', ['&PyTuple_Type', '&py_' + pname],[pname])
-	info.arglist.append(pname)
+        info.varlist.add('GdkRectangle', pname + ' = { 0, 0, 0, 0 }')
+        info.varlist.add('PyObject', '*py_' + pname + ' = Py_None')
+        info.add_parselist('O', ['&py_' + pname], [pname])
+	info.arglist.append('&' + pname)
 	info.codebefore.append(self.null % {'name':  pname})
 
 class GdkRectangleArg(ArgType):
-    def write_return(self, ptype, info):
+    def write_return(self, ptype, ownsreturn, info):
 	info.varlist.add('GdkRectangle', 'ret')
-	info.codeafter.append('    return Py_BuildValue("(iiii)", ret.x, ret.y, ret.width, ret.height);\n')
+	info.codeafter.append('    return pyg_boxed_new(GDK_TYPE_RECTANGLE, &ret, TRUE, TRUE);')
 
-	
 class ArgMatcher:
     def __init__(self):
 	self.argtypes = {}
@@ -540,6 +664,7 @@ class ArgMatcher:
             self.register('GdkBitmap', oa)
             self.register('GdkBitmap*', oa)
     def register_boxed(self, ptype, typecode):
+        if self.argtypes.has_key(ptype): return
         arg = BoxedArg(ptype, typecode)
         self.register(ptype, arg)
 	self.register(ptype+'*', arg)
@@ -548,9 +673,19 @@ class ArgMatcher:
         arg = CustomBoxedArg(ptype, pytype, getter, new)
 	self.register(ptype+'*', arg)
         self.register('const-'+ptype+'*', arg)
+    def register_pointer(self, ptype, typecode):
+        arg = PointerArg(ptype, typecode)
+        self.register(ptype, arg)
+	self.register(ptype+'*', arg)
+        self.register('const-'+ptype+'*', arg)
 
     def get(self, ptype):
-	return self.argtypes[ptype]
+        try:
+            return self.argtypes[ptype]
+        except KeyError:
+            if ptype[:8] == 'GdkEvent' and ptype[-1] == '*':
+                return self.argtypes['GdkEvent*']
+            raise
     def object_is_a(self, otype, parent):
         if otype == None: return 0
         if otype == parent: return 1
@@ -581,6 +716,9 @@ matcher.register('char', arg)
 matcher.register('gchar', arg)
 matcher.register('guchar', arg)
 
+arg = GUniCharArg()
+matcher.register('gunichar', arg)
+
 arg = IntArg()
 matcher.register('int', arg)
 matcher.register('gint', arg)
@@ -590,16 +728,32 @@ matcher.register('gshort', arg)
 matcher.register('gushort', arg)
 matcher.register('long', arg)
 matcher.register('glong', arg)
-matcher.register('gulong', arg)
 matcher.register('gboolean', arg)
 
 matcher.register('guint8', arg)
 matcher.register('gint8', arg)
 matcher.register('guint16', arg)
 matcher.register('gint16', arg)
-matcher.register('guint32', arg)
 matcher.register('gint32', arg)
-matcher.register('GtkType', arg)
+
+# If the system maxint is smaller than unsigned int, we need to use
+# Long objects with PyLong_AsUnsignedLong
+if sys.maxint >= (1L << 32):
+    matcher.register('guint32', arg)
+else:
+    arg = ULongArg()
+    matcher.register('guint32', arg)
+
+arg = ULongArg()
+matcher.register('gulong', arg)
+
+arg = Int64Arg()
+matcher.register('gint64', arg)
+matcher.register('long-long', arg)
+
+arg = UInt64Arg()
+matcher.register('guint64', arg)
+matcher.register('unsigned-long-long', arg)
 
 arg = DoubleArg()
 matcher.register('double', arg)
@@ -620,8 +774,11 @@ matcher.register('GtkType', GTypeArg())
 matcher.register('GError**', GErrorArg())
 matcher.register('GtkTreePath*', GtkTreePathArg())
 matcher.register('GdkRectangle*', GdkRectanglePtrArg())
+matcher.register('GtkAllocation*', GdkRectanglePtrArg())
 matcher.register('GdkRectangle', GdkRectangleArg())
 
+matcher.register('GdkNativeWindow', ULongArg())
+
 matcher.register_object('GObject', None, 'G_TYPE_OBJECT')
 
 del arg
