@@ -178,7 +178,10 @@ static PyTypeObject PyGtkStyleHelper_Type = {
     (hashfunc)0,
     (ternaryfunc)0,
     (reprfunc)0,
-    0L,0L,0L,0L,
+    (getattrofunc)0,
+    (setattrofunc)0,
+    0,
+    Py_TPFLAGS_DEFAULT,
     NULL
 };
 
@@ -677,7 +680,10 @@ PyTypeObject PyGdkWindow_Type = {
     (hashfunc)PyGdkWindow_Hash,
     (ternaryfunc)0,
     (reprfunc)0,
-    0L,0L,0L,0L,
+    (getattrofunc)0,
+    (setattrofunc)0,
+    0,
+    Py_TPFLAGS_DEFAULT,
     NULL
 };
 #endif
@@ -753,10 +759,298 @@ PyTypeObject PyGdkAtom_Type = {
     (hashfunc)pygdk_atom_hash,
     (ternaryfunc)0,
     (reprfunc)pygdk_atom_str,
-    0L,0L,0L,0L,
+    (getattrofunc)0,
+    (setattrofunc)0,
+    0,
+    Py_TPFLAGS_DEFAULT,
     NULL
 };
 
+typedef struct {
+    PyObject_HEAD
+    GtkTreeModel *model;
+    GtkTreeIter iter;
+} PyGtkTreeModelRow;
+staticforward PyTypeObject PyGtkTreeModelRow_Type;
+
+PyObject *
+_pygtk_tree_model_row_new(GtkTreeModel *model, GtkTreeIter *iter)
+{
+    PyGtkTreeModelRow *self;
+
+    self = (PyGtkTreeModelRow *) PyObject_NEW(PyGtkTreeModelRow,
+					      &PyGtkTreeModelRow_Type);
+    if (self == NULL)
+	return NULL;
+    self->model = g_object_ref(model);
+    self->iter = *iter;
+    return (PyObject *)self;
+}
+
+static void
+pygtk_tree_model_row_dealloc(PyGtkTreeModelRow *self)
+{
+    g_object_unref(self->model);
+    PyObject_DEL(self);
+}
+
+static int
+pygtk_tree_model_row_length(PyGtkTreeModelRow *self)
+{
+    return gtk_tree_model_get_n_columns(self->model);
+}
+
+static PyObject *
+pygtk_tree_model_row_getitem(PyGtkTreeModelRow *self, gint column)
+{
+    gint n_columns;
+    GValue value = { 0, };
+    PyObject *ret;
+
+    n_columns = gtk_tree_model_get_n_columns(self->model);
+    if (column < 0)
+	column += n_columns;
+    if (column < 0 || column >= n_columns) {
+	PyErr_SetString(PyExc_IndexError, "column index out of range");
+        return NULL;
+    }
+    gtk_tree_model_get_value(self->model, &self->iter, column, &value);
+    ret = pyg_value_as_pyobject(&value, TRUE);
+    g_value_unset(&value);
+    return ret;
+}
+
+static int
+pygtk_tree_model_row_setitem(PyGtkTreeModelRow *self, gint column,
+			     PyObject *pyvalue)
+{
+    gint n_columns;
+    GValue value = { 0, };
+
+    if (!GTK_IS_LIST_STORE(self->model) && !GTK_IS_TREE_STORE(self->model)) {
+	PyErr_SetString(PyExc_TypeError,
+			"can not set cells in this tree model");
+	return -1;
+    }
+
+    n_columns = gtk_tree_model_get_n_columns(self->model);
+    if (column < 0)
+	column += n_columns;
+    if (column < 0 || column >= n_columns) {
+	PyErr_SetString(PyExc_IndexError, "column index out of range");
+        return -1;
+    }
+    g_value_init(&value, gtk_tree_model_get_column_type(self->model, column));
+    if (pyg_value_from_pyobject(&value, pyvalue)) {
+	PyErr_SetString(PyExc_TypeError,
+			"value is of wrong type for this column");
+	return -1;
+    }
+    if (GTK_IS_LIST_STORE(self->model))
+	gtk_list_store_set_value(GTK_LIST_STORE(self->model), &self->iter,
+				 column, &value);
+    else if (GTK_IS_TREE_STORE(self->model))
+	gtk_tree_store_set_value(GTK_TREE_STORE(self->model), &self->iter,
+				 column, &value);
+    g_value_unset(&value);
+    return 0;
+}
+
+static PySequenceMethods pygtk_tree_model_row_seqmethods = {
+    (inquiry)pygtk_tree_model_row_length,
+    (binaryfunc)0,
+    (intargfunc)0,
+    (intargfunc)pygtk_tree_model_row_getitem,
+    (intintargfunc)0,
+    (intobjargproc)pygtk_tree_model_row_setitem,
+    (intintobjargproc)0
+};
+
+static PyObject *
+pygtk_tree_model_row_iterchildren(PyGtkTreeModelRow *self)
+{
+    return _pygtk_tree_model_row_iter_new(self->model, &self->iter);
+}
+
+static PyMethodDef pygtk_tree_model_row_methods[] = {
+    { "iterchildren", (PyCFunction)pygtk_tree_model_row_iterchildren, METH_NOARGS },
+    { NULL, NULL, 0 }
+};
+
+static PyObject *
+pygtk_tree_model_row_get_next(PyGtkTreeModelRow *self, void *closure)
+{
+    GtkTreeIter iter;
+
+    iter = self->iter;
+    if (gtk_tree_model_iter_next(self->model, &iter))
+	return _pygtk_tree_model_row_new(self->model, &iter);
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject *
+pygtk_tree_model_row_get_parent(PyGtkTreeModelRow *self, void *closure)
+{
+    GtkTreeIter parent;
+
+    if (gtk_tree_model_iter_parent(self->model, &self->iter, &parent))
+	return _pygtk_tree_model_row_new(self->model, &parent);
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject *
+pygtk_tree_model_row_get_model(PyGtkTreeModelRow *self, void *closure)
+{
+    return pygobject_new((GObject *)self->model);
+}
+
+static PyObject *
+pygtk_tree_model_row_get_path(PyGtkTreeModelRow *self, void *closure)
+{
+    GtkTreePath *path;
+    PyObject *ret;
+
+    path = gtk_tree_model_get_path(self->model, &self->iter);
+    if (!path) {
+	PyErr_SetString(PyExc_RuntimeError, "could not get tree path");
+	return NULL;
+    }
+    ret = pygtk_tree_path_to_pyobject(path);
+    gtk_tree_path_free(path);
+    return ret;
+}
+
+static PyGetSetDef pygtk_tree_model_row_getsets[] = {
+    { "next", (getter)pygtk_tree_model_row_get_next, (setter)0 },
+    { "parent", (getter)pygtk_tree_model_row_get_parent, (setter)0 },
+    { "model", (getter)pygtk_tree_model_row_get_model, (setter)0 },
+    { "path", (getter)pygtk_tree_model_row_get_path, (setter)0 },
+    { NULL, (getter)0, (setter)0 }
+};
+
+static PyTypeObject PyGtkTreeModelRow_Type = {
+    PyObject_HEAD_INIT(NULL)
+    0,
+    "gtk.gtk.TreeModelRow",
+    sizeof(PyGtkTreeModelRow),
+    0,
+    (destructor)pygtk_tree_model_row_dealloc,
+    (printfunc)0,
+    (getattrfunc)0,
+    (setattrfunc)0,
+    (cmpfunc)0,
+    (reprfunc)0,
+    0,
+    &pygtk_tree_model_row_seqmethods,
+    0,
+    (hashfunc)0,
+    (ternaryfunc)0,
+    (reprfunc)0,
+    (getattrofunc)0,
+    (setattrofunc)0,
+    0,
+    Py_TPFLAGS_DEFAULT,
+    NULL,
+    (traverseproc)0,
+    (inquiry)0,
+    (richcmpfunc)0,
+    0,
+    (getiterfunc)0,
+    (iternextfunc)0,
+    pygtk_tree_model_row_methods,
+    0,
+    pygtk_tree_model_row_getsets
+};
+
+typedef struct {
+    PyObject_HEAD
+    GtkTreeModel *model;
+    gboolean has_more;
+    GtkTreeIter iter;
+} PyGtkTreeModelRowIter;
+staticforward PyTypeObject PyGtkTreeModelRowIter_Type;
+
+PyObject *
+_pygtk_tree_model_row_iter_new(GtkTreeModel *model, GtkTreeIter *parent_iter)
+{
+    PyGtkTreeModelRowIter *self;
+
+    self = (PyGtkTreeModelRowIter *) PyObject_NEW(PyGtkTreeModelRowIter,
+						  &PyGtkTreeModelRowIter_Type);
+    if (self == NULL)
+	return NULL;
+    self->model = g_object_ref(model);
+    /* iterate through child nodes */
+    self->has_more = gtk_tree_model_iter_children(self->model, &self->iter,
+						  parent_iter);
+    return (PyObject *)self;
+}
+
+static void
+pygtk_tree_model_row_iter_dealloc(PyGtkTreeModelRowIter *self)
+{
+    g_object_unref(self->model);
+    PyObject_DEL(self);
+}
+
+PyObject *
+pygtk_tree_model_row_iter_getiter(PyGtkTreeModelRowIter *self)
+{
+    Py_INCREF(self);
+    return (PyObject *)self;
+}
+
+PyObject *
+pygtk_tree_model_row_iter_next(PyGtkTreeModelRowIter *self)
+{
+    PyObject *row;
+
+    if (!self->has_more) {
+	PyErr_SetNone(PyExc_StopIteration);
+	return NULL;
+    }
+
+    row = _pygtk_tree_model_row_new(self->model, &self->iter);
+
+    /* move to next iter */
+    self->has_more = gtk_tree_model_iter_next(self->model, &self->iter);
+
+    return row;
+}
+
+static PyTypeObject PyGtkTreeModelRowIter_Type = {
+    PyObject_HEAD_INIT(NULL)
+    0,
+    "gtk.gtk.TreeModelRowIter",
+    sizeof(PyGtkTreeModelRowIter),
+    0,
+    (destructor)pygtk_tree_model_row_iter_dealloc,
+    (printfunc)0,
+    (getattrfunc)0,
+    (setattrfunc)0,
+    (cmpfunc)0,
+    (reprfunc)0,
+    0,
+    0,
+    0,
+    (hashfunc)0,
+    (ternaryfunc)0,
+    (reprfunc)0,
+    (getattrofunc)0,
+    (setattrofunc)0,
+    0,
+    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_ITER,
+    NULL,
+    (traverseproc)0,
+    (inquiry)0,
+    (richcmpfunc)0,
+    0,
+    (getiterfunc)pygtk_tree_model_row_iter_getiter,
+    (iternextfunc)pygtk_tree_model_row_iter_next
+};
+
 PyObject *
 pygtk_tree_path_to_pyobject(GtkTreePath *path)
 {
@@ -774,7 +1068,13 @@ pygtk_tree_path_to_pyobject(GtkTreePath *path)
 GtkTreePath *
 pygtk_tree_path_from_pyobject(PyObject *object)
 {
-    if (PyTuple_Check(object)) {
+    if (PyInt_Check(object)) {
+	GtkTreePath *path;
+
+	path = gtk_tree_path_new();
+	gtk_tree_path_append_index(path, PyInt_AsLong(object));
+	return path;
+    } else if (PyTuple_Check(object)) {
 	GtkTreePath *path;
 	guint len, i;
 
@@ -819,22 +1119,65 @@ PyGtkTreePath_to_value(GValue *value, PyObject *object)
     return -1;
 }
 
+gboolean
+pygdk_rectangle_from_pyobject(PyObject *object, GdkRectangle *rectangle)
+{
+    g_return_val_if_fail(rectangle != NULL, FALSE);
+
+    if (pyg_boxed_check(object, GDK_TYPE_RECTANGLE)) {
+	*rectangle = *pyg_boxed_get(object, GdkRectangle);
+	return TRUE;
+    }
+    if (PyArg_ParseTuple(object, "iiii", &rectangle->x, &rectangle->y,
+				&rectangle->width, &rectangle->height)) {
+	return TRUE;
+    }
+    PyErr_Clear();
+    PyErr_SetString(PyExc_TypeError, "could not convert to GdkRectangle");
+    return FALSE;
+}
+
+static PyObject *
+PyGdkRectangle_from_value(const GValue *value)
+{
+    GdkRectangle *rect = (GdkRectangle *)g_value_get_boxed(value);
+
+    return pyg_boxed_new(GDK_TYPE_RECTANGLE, rect, TRUE, TRUE);
+}
+static int
+PyGdkRectangle_to_value(GValue *value, PyObject *object)
+{
+    GdkRectangle rect;
+
+    if (!pygdk_rectangle_from_pyobject(object, &rect))
+	return -1;
+
+    g_value_set_boxed(value, &rect);
+    return 0;
+}
+
 /* We have to set ob_type here because stupid win32 does not allow you
  * to use variables from another dll in a global variable initialisation.
  */
 void
 _pygtk_register_boxed_types(PyObject *moddict)
 {
-#define register_tp(x) Py##x##_Type.ob_type = &PyType_Type; \
-    PyDict_SetItemString(moddict, #x "Type", (PyObject *)&Py##x##_Type)
-
     PyGtkStyleHelper_Type.ob_type = &PyType_Type;
     PyGdkAtom_Type.ob_type = &PyType_Type;
-#if 0
-    register_tp(GdkWindow);
-#endif
-    register_tp(GdkAtom);
+    PyGtkTreeModelRow_Type.ob_type = &PyType_Type;
+    PyGtkTreeModelRowIter_Type.ob_type = &PyType_Type;
+
+    PyType_Ready(&PyGtkStyleHelper_Type);
+    PyType_Ready(&PyGdkAtom_Type);
+    PyType_Ready(&PyGtkTreeModelRow_Type);
+    PyType_Ready(&PyGtkTreeModelRowIter_Type);
+
+    PyDict_SetItemString(moddict, "GdkAtomType", (PyObject *)&PyGdkAtom_Type);
+
     pyg_register_boxed_custom(GTK_TYPE_TREE_PATH,
 			      PyGtkTreePath_from_value,
 			      PyGtkTreePath_to_value);
+    pyg_register_boxed_custom(GDK_TYPE_RECTANGLE,
+			      PyGdkRectangle_from_value,
+			      PyGdkRectangle_to_value);
 }
