@@ -1,3 +1,4 @@
+/* -*- Mode: C; c-basic-offset: 4 -*- */
 #include "pygobject-private.h"
 
 static const gchar *pygobject_class_id     = "PyGObject::class";
@@ -126,6 +127,7 @@ pygobject_new(GObject *obj)
     self->hasref = FALSE;
     self->inst_dict = NULL;
     self->weakreflist = NULL;
+    self->closures = NULL;
     /* save wrapper pointer so we can access it later */
     g_object_set_qdata(obj, pygobject_wrapper_key, self);
 
@@ -134,12 +136,36 @@ pygobject_new(GObject *obj)
     return (PyObject *)self;
 }
 
+static void
+pygobject_unwatch_closure(gpointer data, GClosure *closure)
+{
+    PyGObject *self = (PyGObject *)data;
+
+    self->closures = g_slist_remove (self->closures, closure);
+}
+
+void
+pygobject_watch_closure(PyObject *self, GClosure *closure)
+{
+    PyGObject *gself;
+
+    g_return_if_fail(self != NULL);
+    g_return_if_fail(PyObject_TypeCheck(self, &PyGObject_Type));
+    g_return_if_fail(closure != NULL);
+    g_return_if_fail(g_slist_find(((PyGObject *)self)->closures, closure) == NULL);
+
+    gself = (PyGObject *)self;
+    gself->closures = g_slist_prepend(gself->closures, closure);
+    g_closure_add_invalidate_notifier(closure,self, pygobject_unwatch_closure);
+}
+
 /* -------------- PyGObject behaviour ----------------- */
 
 static void
 pygobject_dealloc(PyGObject *self)
 {
     GObject *obj = self->obj;
+    GSList *tmp;
 
     if (!pygobject_ownedref_key)
 	pygobject_ownedref_key =
@@ -181,6 +207,17 @@ pygobject_dealloc(PyGObject *self)
     }
     self->inst_dict = NULL;
 
+    tmp = self->closures;
+    while (tmp) {
+	GClosure *closure = tmp->data;
+
+	/* we get next item first, because the current link gets
+	 * invalidated by pygobject_unwatch_closure */
+	tmp = tmp->next;
+	g_closure_invalidate(closure);
+    }
+    self->closures = NULL;
+
     /* the following causes problems with subclassed types */
     /*self->ob_type->tp_free((PyObject *)self); */
     PyObject_GC_Del(self);
@@ -216,8 +253,44 @@ pygobject_repr(PyGObject *self)
 static int
 pygobject_traverse(PyGObject *self, visitproc visit, void *arg)
 {
-    if (self->inst_dict)
-	return visit(self->inst_dict, arg);
+    int ret = 0;
+    GSList *tmp;
+
+    if (self->inst_dict) ret = visit(self->inst_dict, arg);
+    if (ret != 0) return ret;
+
+    for (tmp = self->closures; tmp != NULL; tmp = tmp->next) {
+	PyGClosure *closure = tmp->data;
+
+	if (closure->callback) ret = visit(closure->callback, arg);
+	if (ret != 0) return ret;
+
+	if (closure->extra_args) ret = visit(closure->extra_args, arg);
+	if (ret != 0) return ret;
+
+	if (closure->swap_data) ret = visit(closure->swap_data, arg);
+	if (ret != 0) return ret;
+    }
+    return 0;
+}
+
+static int
+pygobject_clear(PyGObject *self)
+{
+    GSList *tmp;
+
+    tmp = self->closures;
+    while (tmp) {
+	GClosure *closure = tmp->data;
+
+	/* we get next item first, because the current link gets
+	 * invalidated by pygobject_unwatch_closure */
+	tmp = tmp->next;
+	g_closure_invalidate(closure);
+    }
+    if (self->closures != NULL)
+	g_message("invalidated all closures, but self->closures != NULL !");
+
     return 0;
 }
 
@@ -227,12 +300,29 @@ pygobject_free(PyObject *op)
     PyObject_GC_Del(op);
 }
 
+
 /* ---------------- PyGObject methods ----------------- */
 
+static void
+parameter_list_free (GParameter *parameters, guint n_parameters)
+{
+	gint i;
+	for (i = 0; i < n_parameters; i ++){
+		g_free ((char *)parameters[i].name);
+		g_value_unset (&(parameters[i].value));
+	}
+	g_free (parameters);
+}
+
 static int
 pygobject_init(PyGObject *self, PyObject *args, PyObject *kwargs)
 {
     GType object_type;
+    guint n_parameters = 0;
+    GParameter *parameters = NULL;
+    PyObject *key, *item;
+    gint pos = 0;
+    GObjectClass *class;
 
     if (!PyArg_ParseTuple(args, ":GObject.__init__", &object_type))
 	return -1;
@@ -241,7 +331,56 @@ pygobject_init(PyGObject *self, PyObject *args, PyObject *kwargs)
     if (!object_type)
 	return -1;
 
-    self->obj = g_object_new(object_type, NULL);
+    if ((class = g_type_class_ref (object_type)) == NULL) {
+	PyErr_SetString(PyExc_TypeError,
+			"could not get a reference to type class");
+	return -1;
+    }
+
+
+    while (kwargs && PyDict_Next(kwargs, &pos, &key, &item)) {
+	    gchar *param_name;
+	    GParamSpec *pspec;
+	    GValue value = { 0, };
+
+	    param_name = PyString_AsString(key);
+	    pspec = g_object_class_find_property(class, param_name);
+
+	    if (pspec == NULL) {
+		    gchar buf[128];
+		    
+		    g_snprintf(buf, sizeof(buf),
+			       "Unknown parameter '%s' used in type initialization arguments",
+			       param_name);
+		    PyErr_SetString(PyExc_AttributeError, buf);
+		    parameter_list_free (parameters, n_parameters);
+		    g_type_class_unref(class);
+		    return -1;
+	    }
+	    g_value_init(&value, G_PARAM_SPEC_VALUE_TYPE(pspec));
+	    if (pyg_value_from_pyobject(&value, item) < 0) {
+		    PyErr_SetString(PyExc_TypeError,
+			"could not convert argument to correct param type");
+		    parameter_list_free (parameters, n_parameters);
+		    g_type_class_unref(class);
+		    return -1;
+	    }
+
+	    n_parameters ++;
+	    if (parameters == NULL) {
+		    parameters = g_new (GParameter, 1);
+	    } else {
+		    parameters = g_renew (GParameter, parameters, n_parameters);
+	    }
+	    parameters[n_parameters - 1].name = g_strdup (param_name);
+	    parameters[n_parameters - 1].value = value;
+	}
+
+    self->obj = g_object_newv(object_type, n_parameters, parameters);
+    parameter_list_free (parameters, n_parameters);
+
+    g_type_class_unref(class);
+
     if (!self->obj) {
 	PyErr_SetString(PyExc_RuntimeError, "could not create object");
 	return -1;
@@ -279,7 +418,7 @@ pygobject_get_property(PyGObject *self, PyObject *args)
     }
     g_value_init(&value, G_PARAM_SPEC_VALUE_TYPE(pspec));
     g_object_get_property(self->obj, param_name, &value);
-    ret = pyg_value_as_pyobject(&value);
+    ret = pyg_value_as_pyobject(&value, TRUE);
     g_value_unset(&value);
     return ret;
 }
@@ -385,6 +524,7 @@ pygobject_connect(PyGObject *self, PyObject *args)
     gchar *name;
     guint handlerid, sigid, len;
     GQuark detail = 0;
+    GClosure *closure;
 
     len = PyTuple_Size(args);
     if (len < 2) {
@@ -410,8 +550,11 @@ pygobject_connect(PyGObject *self, PyObject *args)
     extra_args = PySequence_GetSlice(args, 2, len);
     if (extra_args == NULL)
 	return NULL;
+    closure = pyg_closure_new(callback, extra_args, NULL);
+    pygobject_watch_closure((PyObject *)self, closure);
     handlerid = g_signal_connect_closure_by_id(self->obj, sigid, detail,
-			pyg_closure_new(callback, extra_args, NULL), FALSE);
+					       closure, FALSE);
+    Py_DECREF(extra_args);
     return PyInt_FromLong(handlerid);
 }
 
@@ -422,6 +565,7 @@ pygobject_connect_after(PyGObject *self, PyObject *args)
     gchar *name;
     guint handlerid, sigid, len;
     GQuark detail;
+    GClosure *closure;
 
     len = PyTuple_Size(args);
     if (len < 2) {
@@ -448,8 +592,11 @@ pygobject_connect_after(PyGObject *self, PyObject *args)
     extra_args = PySequence_GetSlice(args, 2, len);
     if (extra_args == NULL)
 	return NULL;
+    closure = pyg_closure_new(callback, extra_args, NULL);
+    pygobject_watch_closure((PyObject *)self, closure);
     handlerid = g_signal_connect_closure_by_id(self->obj, sigid, detail,
-			pyg_closure_new(callback, extra_args, NULL), TRUE);
+					       closure, TRUE);
+    Py_DECREF(extra_args);
     return PyInt_FromLong(handlerid);
 }
 
@@ -460,6 +607,7 @@ pygobject_connect_object(PyGObject *self, PyObject *args)
     gchar *name;
     guint handlerid, sigid, len;
     GQuark detail;
+    GClosure *closure;
 
     len = PyTuple_Size(args);
     if (len < 3) {
@@ -486,8 +634,11 @@ pygobject_connect_object(PyGObject *self, PyObject *args)
     extra_args = PySequence_GetSlice(args, 3, len);
     if (extra_args == NULL)
 	return NULL;
+    closure = pyg_closure_new(callback, extra_args, NULL);
+    pygobject_watch_closure((PyObject *)self, closure);
     handlerid = g_signal_connect_closure_by_id(self->obj, sigid, detail,
-			pyg_closure_new(callback, extra_args, object), FALSE);
+					       closure, FALSE);
+    Py_DECREF(extra_args);
     return PyInt_FromLong(handlerid);
 }
 
@@ -498,6 +649,7 @@ pygobject_connect_object_after(PyGObject *self, PyObject *args)
     gchar *name;
     guint handlerid, sigid, len;
     GQuark detail;
+    GClosure *closure;
 
     len = PyTuple_Size(args);
     if (len < 3) {
@@ -524,8 +676,11 @@ pygobject_connect_object_after(PyGObject *self, PyObject *args)
     extra_args = PySequence_GetSlice(args, 3, len);
     if (extra_args == NULL)
 	return NULL;
+    closure = pyg_closure_new(callback, extra_args, NULL);
+    pygobject_watch_closure((PyObject *)self, closure);
     handlerid = g_signal_connect_closure_by_id(self->obj, sigid, detail,
-			pyg_closure_new(callback, extra_args, object), TRUE);
+					       closure, TRUE);
+    Py_DECREF(extra_args);
     return PyInt_FromLong(handlerid);
 }
 
@@ -633,7 +788,7 @@ pygobject_emit(PyGObject *self, PyObject *args)
 	g_value_unset(&params[i]);
     g_free(params);
     if ((query.return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE) != G_TYPE_NONE) {
-	py_ret = pyg_value_as_pyobject(&ret);
+	py_ret = pyg_value_as_pyobject(&ret, TRUE);
 	g_value_unset(&ret);
     } else {
 	Py_INCREF(Py_None);
@@ -721,7 +876,7 @@ pygobject_chain_from_overridden(PyGObject *self, PyObject *args)
 	g_value_unset(&params[i]);
     g_free(params);
     if ((query.return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE) != G_TYPE_NONE) {
-	py_ret = pyg_value_as_pyobject(&ret);
+	py_ret = pyg_value_as_pyobject(&ret, TRUE);
 	g_value_unset(&ret);
     } else {
 	Py_INCREF(Py_None);
@@ -767,8 +922,15 @@ pygobject_get_dict(PyGObject *self, void *closure)
     return self->inst_dict;
 }
 
+static PyObject *
+pygobject_get_refcount(PyGObject *self, void *closure)
+{
+    return PyInt_FromLong(self->obj->ref_count);
+}
+
 static PyGetSetDef pygobject_getsets[] = {
     { "__dict__", (getter)pygobject_get_dict, (setter)0 },
+    { "__grefcount__", (getter)pygobject_get_refcount, (setter)0, },
     { NULL, 0, 0 }
 };
 
@@ -798,7 +960,7 @@ PyTypeObject PyGObject_Type = {
     	Py_TPFLAGS_HAVE_GC,		/* tp_flags */
     NULL, /* Documentation string */
     (traverseproc)pygobject_traverse,	/* tp_traverse */
-    (inquiry)0,				/* tp_clear */
+    (inquiry)pygobject_clear,		/* tp_clear */
     (richcmpfunc)0,			/* tp_richcompare */
     offsetof(PyGObject, weakreflist),	/* tp_weaklistoffset */
     (getiterfunc)0,			/* tp_iter */
@@ -812,8 +974,8 @@ PyTypeObject PyGObject_Type = {
     0,					/* tp_descr_set */
     offsetof(PyGObject, inst_dict),	/* tp_dictoffset */
     (initproc)pygobject_init,		/* tp_init */
-    PyType_GenericAlloc,		/* tp_alloc */
-    PyType_GenericNew,			/* tp_new */
+    (allocfunc)0,			/* tp_alloc */
+    (newfunc)0,				/* tp_new */
     pygobject_free,			/* tp_free */
     (inquiry)0,				/* tp_is_gc */
     (PyObject *)0,			/* tp_bases */
